<?php

/**
 * @file
 * Rate request/response functions for the Commerce Canada Post module.
 */


/**
 * Returns a SimpleXML Element containing the rate request for an order.
 *
 * @param object $order
 *   The order to generate a rate request for.
 *
 * @return SimpleXMLElement
 *   A SimpleXML Element containing the rate request or FALSE if a rate request
 *   could not be generated.
 */
function commerce_canadapost_build_rate_request($order) {
  $order_wrapper = entity_metadata_wrapper('commerce_order', $order);

  // Determine the shipping profile reference field name for the order.
  $field_name = commerce_physical_order_shipping_field_name($order);
  $shipping_profile = $order_wrapper->{$field_name}->value();

  // Prepare the shipping address for use in the request.
  if (!empty($order_wrapper->{$field_name}->commerce_customer_address)) {
    $shipping_address = $order_wrapper->{$field_name}->commerce_customer_address->value();
  }
  else {
    $shipping_address = array(
      'locality' => '',
      'administrative_area' => '',
      'postal_code' => '',
      'country' => '',
    );
  }

  // Calculate order total (products only) and convert it to CAD.
  $order_total = commerce_line_items_total($order_wrapper->commerce_line_items, array('product'));
  $items_price = commerce_currency_convert($order_total['amount'], $order_total['currency_code'], 'CAD');

  // Create an array of rate request data.
  // The itemsPrice (total amount in $ for insurance calculation) element is
  // optional.
  $rate_request_data = array(
    'language' => 'en',
    'ratesAndServicesRequest' => array(
      'merchantCPCID' => variable_get('commerce_canadapost_cpcid', 'CPC_DEMO_XML'),
      'fromPostalCode' => variable_get('commerce_canadapost_origin', ''),
      'turnAroundTime' => variable_get('commerce_canadapost_turnaround', ''),
      'itemsPrice' => commerce_currency_amount_to_decimal($items_price, 'CAD'),
      'lineItems' => commerce_canadapost_build_package_data($order),
      'city' => $shipping_address['locality'],
      'provOrState' => $shipping_address['administrative_area'],
      'country' => $shipping_address['country'],
      'postalCode' => $shipping_address['postal_code'],
    ),
  );

  // Remove optional settings if empty.
  if ($rate_request_data['ratesAndServicesRequest']['fromPostalCode'] == '') {
    unset($rate_request_data['ratesAndServicesRequest']['fromPostalCode']);
  }
  if ($rate_request_data['ratesAndServicesRequest']['turnAroundTime'] == '') {
    unset($rate_request_data['ratesAndServicesRequest']['turnAroundTime']);
  }

  // Build and populate the rate request SimpleXML element.
  $rate_request_element = new SimpleXMLElement('<eparcel/>');
  commerce_simplexml_add_children($rate_request_element, $rate_request_data);

  // Allow modules an opportunity to alter the request before it is sent.
  drupal_alter('commerce_canadapost_build_rate_request', $rate_request_element, $order);

  return $rate_request_element;
}

/**
 * Builds array of package data for rate request.
 *
 * The data created by this function does not correspond to any real packages.
 * Since the rates from Canada Post are based solely on weight (unless a package
 * is considered light for its size), the module creates virtual packages to get
 * a rate quote.
 *
 * Dimensions of each virtual package are set to 1x1x1 cm and an element called
 * readyToShip is added that tells Canada Post not to use their automatic boxing
 * tool.
 *
 * If a package exceeds the weight limit the order will be divided into multiple
 * virtual packages.
 *
 * @param object $order
 *   The order object to build the package data from.
 *
 * @return array
 *   An associative array containing a list of packages and their attributes
 *   (weight, dimensions, description, etc.)
 */
function commerce_canadapost_build_package_data($order) {
  // Get the total order weight and the weight markup (in kg).
  $order_weight = commerce_physical_order_weight($order, 'kg');
  $weight_markup = physical_weight_convert(variable_get('commerce_canadapost_weight_markup', array('weight' => 0, 'unit' => 'kg')), 'kg');

  // Calculate how many readyToShip packages will be required for the order.
  $num_packages = commerce_canadapost_package_count($order_weight, $weight_markup);

  // Build package data.
  $package_data = array();

  for ($i = 0; $i < $num_packages; $i++) {
    // All values are manditory with the exception of readyToShip.
    $package_data[] = array(
      'item' => array(
        'quantity' => 1,
        'weight' => round($weight_markup['weight'] + ($order_weight['weight'] / $num_packages), 3),
        'length' => 1,
        'width' => 1,
        'height' => 1,
        'description' => 'Package ' . $i,
        'readyToShip' => NULL,
      ),
    );
  }

  return $package_data;
}

/**
 * Submits an API request to the Canada Post XML Processor.
 *
 * @param string $xml
 *   An XML string to submit.
 * @param string $message
 *   (Optional) Log message to use for logged API requests.
 *
 * @return SimpleXMLElement
 *   Returns a SimpleXML Element containing the response from the API server or
 *   FALSE if there was no response or an error occurred.
 */
function commerce_canadapost_api_request($xml, $message = '') {
  // Log the API request if specified.
  $logging = variable_get('commerce_canadapost_log', array());
  if ($logging['request']) {
    if (empty($message)) {
      $message = t('Submitting API request to the Canada Post XML processor');
    }

    watchdog('canadapost', '@message:<pre>@xml</pre>', array('@message' => $message, '@xml' => $xml));
  }

  // Build the array of header information for the request.
  $header = array();
  $header[] = 'Content-type: text/xml; charset=utf-8';
  $header[] = 'Content-length: ' . drupal_strlen($xml);
  $header[] = 'User-Agent: Drupal Commerce Canada Post Shipping module)';

  // POST the XML to the XML Processor.
  $ch = curl_init('http://sellonline.canadapost.ca:30000/');
  curl_setopt($ch, CURLOPT_HTTPHEADER, $header);
  curl_setopt($ch, CURLOPT_VERBOSE, 0);
  curl_setopt($ch, CURLOPT_POST, 1);
  curl_setopt($ch, CURLOPT_POSTFIELDS, $xml);
  curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
  curl_setopt($ch, CURLOPT_NOPROGRESS, 1);
  curl_setopt($ch, CURLOPT_FOLLOWLOCATION, 0);
  $result = curl_exec($ch);

  // Log any errors to the watchdog.
  if ($error = curl_error($ch)) {
    watchdog('canadapost', 'cURL error: @error', array('@error' => $error), WATCHDOG_ERROR);
    return FALSE;
  }
  curl_close($ch);

  // If we received data back from the server.
  if (!empty($result)) {
    // Extract the result into an XML response object.
    $response = new SimpleXMLElement($result);

    // Log the API request if specified.
    if ($logging['response']) {
      watchdog('canadapost', 'API response received:<pre>@xml</pre>', array('@xml' => $response->asXML()));
    }

    return $response;
  }
  else {
    return FALSE;
  }
}

/**
 * Parse a valid XML response into a commerce_shipping rates array.
 *
 * @param SimpleXMLElement $response
 *   A SimpleXML Element containing the response from the API server.
 *
 * @return array
 *   A commerce_shipping rates array indexed by shipping service or an empty
 *   array if there was a problem parsing the data.
 */
function commerce_canadapost_parse_rate_response($response) {
  $rates = array();
  $service_names = commerce_canadapost_service_names();

  // Count boxes.
  $package_count = 0;
  foreach ($response->ratesAndServicesResponse->packing->box as $package) {
    $package_count++;
  }

  // Build rates array.
  foreach ($response->ratesAndServicesResponse->product as $product) {
    $amount = (float) $product->rate;
    $currency_code = 'CAD';
    $product_id = (string) $product->attributes()->id;

    // Skip if this product id is not in the list of services.
    if (!isset($service_names[$product_id])) {
      continue;
    }

    // Add an item to the rates array for the current service.
    $rates[$service_names[$product_id]] = array(
      'amount' => commerce_currency_decimal_to_amount($amount, $currency_code),
      'currency_code' => $currency_code,
      'data' => array(
        'product_id'            => $product_id,
        'product_name'          => (string) $product->name,
        'shipping_date'         => (string) $product->shippingDate,
        'delivery_date'         => (string) $product->deliveryDate,
        'delivery_dayofweek'    => (string) $product->deliveryDayOfWeek,
        'nextdayam'             => (string) $product->nextDayAM,
        'insurance'             => (string) $response->ratesAndServicesResponse->shippingOptions->insurance,
        'delivery_confirmation' => (string) $response->ratesAndServicesResponse->shippingOptions->deliveryConfirmation,
        'signature'             => (string) $response->ratesAndServicesResponse->shippingOptions->signature,
        'default_markup'        => (float) $response->ratesAndServicesResponse->handling,
        'package_count'         => $package_count,
        'comment'               => (string) $response->ratesAndServicesResponse->comment,
      ),
    );
  }

  return $rates;
}

/**
 * Returns the number of packages required for an order.
 *
 * @param array $order_weight
 *   The weight of the entire order. An associative array with 'weight' and
 *   'unit' keys (see 'physical' module).
 * @param array $weight_markup
 *   (Optional) The weight to add to each package in addition to products. An
 *   associative array with 'weight' and 'unit' keys (see 'physical' module).
 *   Defaults to 0 kilograms.
 * @param array $max_weight
 *   (Optional) Maximum weight for each package. An associative array with
 *   'weight' and 'unit' keys (see 'physical' module). Defaults to 30 kilograms
 *   (the maximum weight allowed by Canada Post).
 *
 * @return int
 *   An integer value specifying the number of packages that will be required to
 *   ship the entire order.
 */
function commerce_canadapost_package_count($order_weight, $weight_markup = array('weight' => 0, 'unit' => 'kg'), $max_weight = array('weight' => 30, 'unit' => 'kg')) {
  // Convert weights to same unit (kilograms).
  $order_weight = physical_weight_convert($order_weight, 'kg');
  $weight_markup = physical_weight_convert($weight_markup, 'kg');

  return ceil($order_weight['weight'] / ($max_weight['weight'] - $weight_markup['weight']));
}
